/*
 * $Id: XIncludeTask.java,v 1.3 2003/09/17 09:56:53 znerd Exp $
 */
package org.znerd.xincludetask;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.Project;
import org.apache.tools.ant.Task;

import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.XMLReader;
import org.xml.sax.helpers.XMLReaderFactory;

import com.elharo.xml.xinclude.SAXXIncluder;
import com.elharo.xml.xinclude.XIncludeFilter;

/**
 * XInclude task.
 *
 * @version $Revision: 1.3 $ $Date: 2003/09/17 09:56:53 $
 * @author <a href="mailto:znerd@FreeBSD.org">Ernst de Haan</a>
 *
 * @see http://www.w3.org/TR/xinclude/
 * @see http://www.ibiblio.org/xml/XInclude/
 */
public class XIncludeTask extends Task {

    /**
     * The name of the input file. This field is set by
     * {@link #setIn(String)}. This field is initially <code>null</code>.
     */
    private String _in;

    /**
     * The name of the output file encoding. This field is set by
     * {@link #setOutEncoding(String)}. This field is initially
     * <code>null</code>.
     */
    private String _outEncoding;

    /**
     * The name of the output file. This field is set by
     * {@link #setOut(String)}.
     */
    private String _out;

    public void execute() throws BuildException {

        // See if the input and output file properties are set
        if ((_in == null) && (_out == null)) {
            throw new BuildException("Neither the \"in\" nor the \"out\" properties are set, while both are mandatory.");
        } else if (_in == null) {
            throw new BuildException("The \"in\" property is not set, although it is mandatory.");
        } else if (_out == null) {
            throw new BuildException("The \"out\" property is not set, although it is mandatory.");
        }

        // Create File instances
        File in  = project.resolveFile(_in);
        File out = project.resolveFile(_out);

        // See if the input file exists
        String inAbs = in.getAbsolutePath();
        if (in.exists() == false) {
            throw new BuildException("The file specified in the \"in\" property (\"" + inAbs + "\") does not exist.");
        } else {
            log("Input file \"" + inAbs + "\" exists.", Project.MSG_DEBUG);
        }

        // See if we can open the file for writing
        String outAbs = out.getAbsolutePath();
        if (out.exists()) {
            if (out.canWrite()) {
                log("Output file \"" + outAbs + "\" exists and is writable.", Project.MSG_DEBUG);
            } else {
                throw new BuildException("Output file \"" + outAbs + "\" exists and is not writable.");
            }
        } else {
            try {
                out.createNewFile();
            } catch (IOException ioException) {
                throw new BuildException("Unable to create output file \"" + outAbs + "\".", ioException);
            } catch (SecurityException securityException) {
                throw new BuildException("Not allowed by security manager to create output file \"" + outAbs + "\".", securityException);
            }
        }

        // Open a stream to the output file
        OutputStream stream;
        try {
            stream = new FileOutputStream(out);
        } catch (IOException ioException) {
            throw new BuildException("Unable to create stream to output file \"" + outAbs + "\".", ioException);
        }

        // Create an XML reader
        XMLReader parser; 
        try {
            parser = XMLReaderFactory.createXMLReader();
            log("Created SAX2 parser: " + parser.getClass().getName(), Project.MSG_DEBUG);
        } catch (SAXException exception) {
            final String defaultDriver = "org.apache.xerces.parsers.SAXParser";
            log("Unable to create a SAX2 parser. Trying default: " + defaultDriver, Project.MSG_DEBUG);
            try {
                parser = XMLReaderFactory.createXMLReader(defaultDriver);
                log("Created SAX2 parser: " + parser.getClass().getName(), Project.MSG_DEBUG);
            } catch (Exception exception2) {
                throw new BuildException("No SAX2 parser available.", exception);
            }
        }
          
        // XXX: Need better namespace handling
        final String feature = "namespace-prefixes";
        try {
            parser.setFeature("http://xml.org/sax/features/" + feature, true);
            log("Successfully set SAX feature \"" + feature + "\".", Project.MSG_DEBUG);
        } catch (SAXException e) {
            throw new BuildException("Unable to set SAX feature \"" + feature + "\".");
        }
          
        // Create an include filter
        XIncludeFilter filter = new XIncludeFilter(); 
        filter.setParent(parser);

        // Determine the output encoding
        String outEncoding;
        if (_outEncoding == null || "".equals(_outEncoding.trim())) {
            outEncoding = "UTF-8";
        } else {
            outEncoding = _outEncoding;
        }

        // Create an XInclude processor
        SAXXIncluder s;
        try {
            log("Creating XInclude processor, output encoding is: " + outEncoding, Project.MSG_VERBOSE);
            s = new SAXXIncluder(stream, outEncoding);
        } catch (UnsupportedEncodingException uee) {
            throw new BuildException("The encoding " + outEncoding + " is not supported.", uee);
        }
        filter.setContentHandler(s);
        final String property = "lexical-handler";
        try {
            filter.setProperty("http://xml.org/sax/properties/" + property, s);
            s.setFilter(filter);
        } catch (SAXException e) {
            log("Failed to set property \"" + property + "\". Ignoring.", Project.MSG_DEBUG);
        }

        // Perform the XInclude processing
        try {
            log("Processing " + inAbs + " to " + outAbs);
            filter.parse(inAbs);
        } catch (SAXException saxException) {
            Exception cause = saxException.getException();
            if (cause == null) {
                cause = saxException;
            }
            if (cause instanceof FileNotFoundException) {
                String filename = cause.getMessage();
                if (filename != null && ("".equals(filename.trim()) == false)) {
                    int i = filename.indexOf(" (No such file or directory)");
                    if (i > 0) {
                       filename = filename.substring(0, i);
                    }
                    throw new BuildException("File not found: " + filename, cause);
                }
            }

            throw new BuildException("Failed to process file: " + inAbs, cause);
        } catch (IOException ioException) {
            throw new BuildException("Failed to process file: " + inAbs, ioException);
        }
    }

    /**
     * Sets the input file.
     *
     * @param in
     *     the input file.
     */
    public void setIn(String in) {
        log("Set input file to: " + in, Project.MSG_DEBUG);
        _in = in;
    }

    /**
     * Sets the output file.
     *
     * @param out
     *     the output file.
     */
    public void setOut(String out) {
        log("Set output file to: " + out, Project.MSG_DEBUG);
        _out = out;
    }

    /**
     * Sets the encoding to use for the output. If the encoding is unsupported,
     * then the build (see {@link #execute()}) will fail.
     *
     * @param encoding
     *     the encoding to use.
     */
    public void setOutEncoding(String encoding) {
        log("Set output encoding to: " + encoding, Project.MSG_DEBUG);
        _outEncoding = encoding;
    }
}
